# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gateway_package.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# This patch in the import was manually written.
# TODO: research the solution to relative imports and fix this hack.
import txp.common.protos.gateway_config_pb2 as gateway__config__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='gateway_package.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x15gateway_package.proto\x1a\x14gateway_config.proto\".\n\x1b\x44imensionSignalSamplesProto\x12\x0f\n\x07samples\x18\x01 \x03(\x02\"\xd8\x01\n\x0bSignalProto\x12-\n\x07samples\x18\x01 \x03(\x0b\x32\x1c.DimensionSignalSamplesProto\x12\x17\n\x0fperception_name\x18\x02 \x01(\t\x12\x1d\n\x15perception_dimensions\x18\x03 \x03(\x05\x12\x1d\n\x10signal_frequency\x18\x04 \x01(\x02H\x00\x88\x01\x01\x12\x1b\n\x13sampling_resolution\x18\x05 \x01(\x05\x12\x11\n\ttimestamp\x18\x06 \x01(\x03\x42\x13\n\x11_signal_frequency\"\xbf\x01\n\x1bGatewayPackageMetadataProto\x12-\n\x0f\x65\x64ge_descriptor\x18\x01 \x01(\x0b\x32\x14.EdgeDescriptorProto\x12-\n\x0fsampling_window\x18\x02 \x01(\x0b\x32\x14.SamplingWindowProto\x12\x12\n\npackage_id\x18\x03 \x01(\t\x12\x1b\n\x13previous_part_index\x18\x04 \x01(\x05\x12\x11\n\ttenant_id\x18\x0b \x01(\t\"~\n\x13GatewayPackageProto\x12.\n\x08metadata\x18\x01 \x01(\x0b\x32\x1c.GatewayPackageMetadataProto\x12\x1d\n\x07signals\x18\x02 \x03(\x0b\x32\x0c.SignalProto\x12\x18\n\x10\x63onfiguration_id\x18\x03 \x01(\tb\x06proto3'
  ,
  dependencies=[gateway__config__pb2.DESCRIPTOR,])




_DIMENSIONSIGNALSAMPLESPROTO = _descriptor.Descriptor(
  name='DimensionSignalSamplesProto',
  full_name='DimensionSignalSamplesProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='samples', full_name='DimensionSignalSamplesProto.samples', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=47,
  serialized_end=93,
)


_SIGNALPROTO = _descriptor.Descriptor(
  name='SignalProto',
  full_name='SignalProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='samples', full_name='SignalProto.samples', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='perception_name', full_name='SignalProto.perception_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='perception_dimensions', full_name='SignalProto.perception_dimensions', index=2,
      number=3, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signal_frequency', full_name='SignalProto.signal_frequency', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sampling_resolution', full_name='SignalProto.sampling_resolution', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='SignalProto.timestamp', index=5,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_signal_frequency', full_name='SignalProto._signal_frequency',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=96,
  serialized_end=312,
)


_GATEWAYPACKAGEMETADATAPROTO = _descriptor.Descriptor(
  name='GatewayPackageMetadataProto',
  full_name='GatewayPackageMetadataProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='edge_descriptor', full_name='GatewayPackageMetadataProto.edge_descriptor', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sampling_window', full_name='GatewayPackageMetadataProto.sampling_window', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='package_id', full_name='GatewayPackageMetadataProto.package_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='previous_part_index', full_name='GatewayPackageMetadataProto.previous_part_index', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tenant_id', full_name='GatewayPackageMetadataProto.tenant_id', index=4,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=315,
  serialized_end=506,
)


_GATEWAYPACKAGEPROTO = _descriptor.Descriptor(
  name='GatewayPackageProto',
  full_name='GatewayPackageProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='GatewayPackageProto.metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signals', full_name='GatewayPackageProto.signals', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='configuration_id', full_name='GatewayPackageProto.configuration_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=508,
  serialized_end=634,
)

_SIGNALPROTO.fields_by_name['samples'].message_type = _DIMENSIONSIGNALSAMPLESPROTO
_SIGNALPROTO.oneofs_by_name['_signal_frequency'].fields.append(
  _SIGNALPROTO.fields_by_name['signal_frequency'])
_SIGNALPROTO.fields_by_name['signal_frequency'].containing_oneof = _SIGNALPROTO.oneofs_by_name['_signal_frequency']
_GATEWAYPACKAGEMETADATAPROTO.fields_by_name['edge_descriptor'].message_type = gateway__config__pb2._EDGEDESCRIPTORPROTO
_GATEWAYPACKAGEMETADATAPROTO.fields_by_name['sampling_window'].message_type = gateway__config__pb2._SAMPLINGWINDOWPROTO
_GATEWAYPACKAGEPROTO.fields_by_name['metadata'].message_type = _GATEWAYPACKAGEMETADATAPROTO
_GATEWAYPACKAGEPROTO.fields_by_name['signals'].message_type = _SIGNALPROTO
DESCRIPTOR.message_types_by_name['DimensionSignalSamplesProto'] = _DIMENSIONSIGNALSAMPLESPROTO
DESCRIPTOR.message_types_by_name['SignalProto'] = _SIGNALPROTO
DESCRIPTOR.message_types_by_name['GatewayPackageMetadataProto'] = _GATEWAYPACKAGEMETADATAPROTO
DESCRIPTOR.message_types_by_name['GatewayPackageProto'] = _GATEWAYPACKAGEPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DimensionSignalSamplesProto = _reflection.GeneratedProtocolMessageType('DimensionSignalSamplesProto', (_message.Message,), {
  'DESCRIPTOR' : _DIMENSIONSIGNALSAMPLESPROTO,
  '__module__' : 'gateway_package_pb2'
  # @@protoc_insertion_point(class_scope:DimensionSignalSamplesProto)
  })
_sym_db.RegisterMessage(DimensionSignalSamplesProto)

SignalProto = _reflection.GeneratedProtocolMessageType('SignalProto', (_message.Message,), {
  'DESCRIPTOR' : _SIGNALPROTO,
  '__module__' : 'gateway_package_pb2'
  # @@protoc_insertion_point(class_scope:SignalProto)
  })
_sym_db.RegisterMessage(SignalProto)

GatewayPackageMetadataProto = _reflection.GeneratedProtocolMessageType('GatewayPackageMetadataProto', (_message.Message,), {
  'DESCRIPTOR' : _GATEWAYPACKAGEMETADATAPROTO,
  '__module__' : 'gateway_package_pb2'
  # @@protoc_insertion_point(class_scope:GatewayPackageMetadataProto)
  })
_sym_db.RegisterMessage(GatewayPackageMetadataProto)

GatewayPackageProto = _reflection.GeneratedProtocolMessageType('GatewayPackageProto', (_message.Message,), {
  'DESCRIPTOR' : _GATEWAYPACKAGEPROTO,
  '__module__' : 'gateway_package_pb2'
  # @@protoc_insertion_point(class_scope:GatewayPackageProto)
  })
_sym_db.RegisterMessage(GatewayPackageProto)


# @@protoc_insertion_point(module_scope)
