# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gateway_config.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='gateway_config.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x14gateway_config.proto\x1a\x1cgoogle/protobuf/struct.proto\"\xd7\x01\n\x13SamplingWindowProto\x12\x18\n\x10observation_time\x18\x01 \x01(\x05\x12\x15\n\rsampling_time\x18\x02 \x01(\x05\x12\x14\n\x0cmachines_ids\x18\x03 \x03(\t\x12\x1d\n\x15observation_timestamp\x18\x04 \x01(\x03\x12\x17\n\x0fgateway_task_id\x18\x05 \x01(\x03\x12\x1d\n\x15sampling_window_index\x18\x06 \x01(\r\x12\"\n\x1anumber_of_sampling_windows\x18\x07 \x01(\r\"b\n\x14MachineMetadataProto\x12\x12\n\nmachine_id\x18\x01 \x01(\t\x12\r\n\x05model\x18\x02 \x01(\t\x12\x14\n\x0cmanufacturer\x18\x03 \x01(\t\x12\x11\n\tedges_ids\x18\x04 \x03(\t\"\x96\x01\n\x13\x45\x64geDescriptorProto\x12\x12\n\nlogical_id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65vice_kind\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65vice_type\x18\x03 \x01(\t\x12\x31\n\x0b\x65\x64ge_params\x18\x04 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x88\x01\x01\x42\x0e\n\x0c_edge_params\"}\n\x12GatewayConfigProto\x12\x12\n\ngateway_id\x18\x01 \x01(\t\x12\x13\n\x0bregistry_id\x18\x02 \x01(\t\x12\x12\n\nproject_id\x18\x03 \x01(\t\x12\x14\n\x0c\x63loud_region\x18\x04 \x01(\t\x12\x14\n\x0cmachines_ids\x18\x06 \x03(\tb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])




_SAMPLINGWINDOWPROTO = _descriptor.Descriptor(
  name='SamplingWindowProto',
  full_name='SamplingWindowProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='observation_time', full_name='SamplingWindowProto.observation_time', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sampling_time', full_name='SamplingWindowProto.sampling_time', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='machines_ids', full_name='SamplingWindowProto.machines_ids', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='observation_timestamp', full_name='SamplingWindowProto.observation_timestamp', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gateway_task_id', full_name='SamplingWindowProto.gateway_task_id', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sampling_window_index', full_name='SamplingWindowProto.sampling_window_index', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='number_of_sampling_windows', full_name='SamplingWindowProto.number_of_sampling_windows', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=55,
  serialized_end=270,
)


_MACHINEMETADATAPROTO = _descriptor.Descriptor(
  name='MachineMetadataProto',
  full_name='MachineMetadataProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='machine_id', full_name='MachineMetadataProto.machine_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model', full_name='MachineMetadataProto.model', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='manufacturer', full_name='MachineMetadataProto.manufacturer', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='edges_ids', full_name='MachineMetadataProto.edges_ids', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=272,
  serialized_end=370,
)


_EDGEDESCRIPTORPROTO = _descriptor.Descriptor(
  name='EdgeDescriptorProto',
  full_name='EdgeDescriptorProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='logical_id', full_name='EdgeDescriptorProto.logical_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='device_kind', full_name='EdgeDescriptorProto.device_kind', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='device_type', full_name='EdgeDescriptorProto.device_type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='edge_params', full_name='EdgeDescriptorProto.edge_params', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_edge_params', full_name='EdgeDescriptorProto._edge_params',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=373,
  serialized_end=523,
)


_GATEWAYCONFIGPROTO = _descriptor.Descriptor(
  name='GatewayConfigProto',
  full_name='GatewayConfigProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='gateway_id', full_name='GatewayConfigProto.gateway_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='registry_id', full_name='GatewayConfigProto.registry_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_id', full_name='GatewayConfigProto.project_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cloud_region', full_name='GatewayConfigProto.cloud_region', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='machines_ids', full_name='GatewayConfigProto.machines_ids', index=4,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=525,
  serialized_end=650,
)

_EDGEDESCRIPTORPROTO.fields_by_name['edge_params'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_EDGEDESCRIPTORPROTO.oneofs_by_name['_edge_params'].fields.append(
  _EDGEDESCRIPTORPROTO.fields_by_name['edge_params'])
_EDGEDESCRIPTORPROTO.fields_by_name['edge_params'].containing_oneof = _EDGEDESCRIPTORPROTO.oneofs_by_name['_edge_params']
DESCRIPTOR.message_types_by_name['SamplingWindowProto'] = _SAMPLINGWINDOWPROTO
DESCRIPTOR.message_types_by_name['MachineMetadataProto'] = _MACHINEMETADATAPROTO
DESCRIPTOR.message_types_by_name['EdgeDescriptorProto'] = _EDGEDESCRIPTORPROTO
DESCRIPTOR.message_types_by_name['GatewayConfigProto'] = _GATEWAYCONFIGPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SamplingWindowProto = _reflection.GeneratedProtocolMessageType('SamplingWindowProto', (_message.Message,), {
  'DESCRIPTOR' : _SAMPLINGWINDOWPROTO,
  '__module__' : 'gateway_config_pb2'
  # @@protoc_insertion_point(class_scope:SamplingWindowProto)
  })
_sym_db.RegisterMessage(SamplingWindowProto)

MachineMetadataProto = _reflection.GeneratedProtocolMessageType('MachineMetadataProto', (_message.Message,), {
  'DESCRIPTOR' : _MACHINEMETADATAPROTO,
  '__module__' : 'gateway_config_pb2'
  # @@protoc_insertion_point(class_scope:MachineMetadataProto)
  })
_sym_db.RegisterMessage(MachineMetadataProto)

EdgeDescriptorProto = _reflection.GeneratedProtocolMessageType('EdgeDescriptorProto', (_message.Message,), {
  'DESCRIPTOR' : _EDGEDESCRIPTORPROTO,
  '__module__' : 'gateway_config_pb2'
  # @@protoc_insertion_point(class_scope:EdgeDescriptorProto)
  })
_sym_db.RegisterMessage(EdgeDescriptorProto)

GatewayConfigProto = _reflection.GeneratedProtocolMessageType('GatewayConfigProto', (_message.Message,), {
  'DESCRIPTOR' : _GATEWAYCONFIGPROTO,
  '__module__' : 'gateway_config_pb2'
  # @@protoc_insertion_point(class_scope:GatewayConfigProto)
  })
_sym_db.RegisterMessage(GatewayConfigProto)


# @@protoc_insertion_point(module_scope)
